# -*- coding: utf-8 -*-

from __future__ import division

import time

from hysys_stream import HysysStream
from exc import MethodFailedError

ETA_GUESS = 0.7 # first attempt to guess the polytropic eff. [%]
MAX_ITERATIONS = 100 # maxmimum number of iterations allowed
TOL = 10e-4 # tolerance error for each step.

TOTAL_NUMBER = 0

def pressure_steps(p1, p2, N):
	"""Generator to make the pressure steps for the given
	range and steps. For every step it returns the pressure start and
	pressure stop.

	E.g.: pressure_steps(1, 2, 2) ->
		(1, 1.5)
		(1.5, 2)

	:param p1: start pressure.
	:param p2: end pressure.
	:param N: number of steps.

	"""
	r = p2 - p1
	if r < 0:
		raise ValueError("p2 {0} must be larger than p1 {1}.".format(p2, p1))

	for i in range(N):
		yield p1 + (r/N)*i, p1 + (r/N)*(i + 1)


class DirectIntegration(object):
	"""Class implementing the direct integration model
	for compressor performance.
	"""

	def __init__(self, p_in, T_in, p_out, T_out, components, N = 10, eos_type = "pengrob"):
		"""Constructor for the DirectIntegration class.

		:param p_in: the inlet (suction) pressure in kPa.
		:param T_in: the inlet (suction) temperature in C.
		:param p_out: the outlet (pressure) pressure in kPa.
		:param T_out: the outlet (pressure) temperature in C.
		:param N: number of steps to divided the compression path in.
		:param components: a dictionary of components and their mole fractions.
		:param eos_type: the eos type, e.g.: "pengrob", "BWRS", etc.

		"""
		# store the input values as member variables.
		self.p_in = p_in
		self.T_in = T_in
		self.p_out = p_out
		self.T_out = T_out
		self.N = N
		self.components = components
		self.eos_type = eos_type

		# initiate the eos for the given components
		self.eos = HysysStream(self.components, self.eos_type)

	def set_number_of_iterations(self, N):
		"""Set a new value for number of steps in the model.

		:param N: new number of steps.

		"""
		self.N = N

	def set_inlet_outlet_conditions(self, p_in, T_in, p_out, T_out):
		"""Set new values for the initial values for inlet and outlet temperature
		and pressure.

		:param p_in: new inlet pressure in kPa.
		:param T_in: new inlet temperature in degrees C.
		:param p_out: new outlet pressure in kPa.
		:param T_out: new outlet temperature in degrees C.

		"""
		self.p_in = p_in
		self.T_in = T_in
		self.p_out = p_out
		self.T_out = T_out

	def calculate_outlet_temp(self, eta):
		"""Calculates the outlet temperature with the direct integration
		model, with a polytropic efficiency equal eta.

		:param eta: the polytopic efficiency (between [0-1])
		:returns: the temperature in C calculated with DI and sum of all enthalpy diffs

		"""
		global TOTAL_NUMBER
		# initiate the iteration process by setting the T1 value to the inlet temperature
		T1 = self.T_in
		delta_h = 0

		result = [[T1, self.p_in]]

		# iterating through all the pressure steps (from p_in to p_out) in N steps.
		for p1, p2 in pressure_steps(self.p_in, self.p_out, self.N):
			# making the temp range for this iteration
			# guessing that the next T is between T1 and T_out*1.2
			T2_low = T1
			# need to expand the range (T_out*1.2) so that T2_high - T2_low doesn't stop the last iterations
			T2_high = self.T_out * 1.2
			
			n = 1 # counter, prevent infinite loop
			while True:
				# make a middle point of T2_low and T2_high
				T2_mid = (T2_low + T2_high)/2
			
				# check if the di_model equal 0 (almost equal) or that the difference in T2_high and T2_low is not too small
				if abs(self.di_model(p1, T1, p2, T2_mid, eta)) < TOL or abs((T2_high - T2_low)/2) < TOL:
					# correct T2 found. break the while loop
					break

				# update the middle point dependent on the last T2_mid "guess"
				if cmp(self.di_model(p1, T1, p2, T2_mid, eta), 0) > 0:
					T2_low = T2_mid
				else:
					T2_high = T2_mid

				
				# increment counter
				n += 1
				TOTAL_NUMBER += 1
				if n > MAX_ITERATIONS:
					raise MethodFailedError("Max iterations reached without finding the outlet temperature within the given tolerances.")

			delta_h += (self.eos.enthalpy(p2, T2_mid) - self.eos.enthalpy(p1, T1))

			result.append([T2_mid, p2])

			# update the temperature value.
			T1 = T2_mid # new T1 found from the itetation
		
		return T1, delta_h, result

	def calculate_correct_eta(self):
		"""Calculates the polytropic efficiency by performing the DI model until the outlet temperature
		corresponds to self.T_out.

		:returns: the correct polytropic efficiency.
		
		"""
		global TOTAL_NUMBER
		# guessing the eta is somewhere between plus minus 20% of the ETA_GUESS variable
		eta_low = 0 # ETA_GUESS*0.8
		eta_high = 1 # ETA_GUESS*1.2
		T = None

		n = 1 # counter, prevent infinite loop
		while 1:
			eta_mid = (eta_low + eta_high)/2

			T, delta_h, result = self.calculate_outlet_temp(eta_mid)

			if abs(T - self.T_out) < TOL:
				break

			if cmp(T, self.T_out) > 0:
				eta_low = eta_mid
			else:
				eta_high = eta_mid

			n += 1
			TOTAL_NUMBER += 1
			if n > MAX_ITERATIONS:
				raise MethodFailedError("Max iterations reached without finding the correct polytropic efficiency within the given tolerances. {0}".format(eta_mid))
		print(TOTAL_NUMBER)
		return eta_mid

	def di_model(self, p1, T1, p2, T2, eta):
		"""The direct integration model implemented as the difference in
		vdp and eta*dh.

		:param p1: inlet pressure in kPa.
		:param T1: inlet temperature in degrees C.
		:param p2: outlet pressure in kPa.
		:param T2: outlet temperature in degrees C.
		:param eta: the polytropic efficiency.
		:returns: the difference in vdp and eta*dh.

		"""
		# extract thremodynamic data drom the HysysStream.
		v1 = self.eos.specific_volume(p1, T1)
		h1 = self.eos.enthalpy(p1, T1)
		v2 = self.eos.specific_volume(p2, T2)
		h2 = self.eos.enthalpy(p2, T2)

		# return the diffrence between v*dp and eta*dh
		return ((v2 + v1)/2)*(p2 - p1) - (h2 - h1)*eta

if __name__ == '__main__':
	# components = {
	# 				"Air": 1,
	# 				# "Ethane": 0.7,
	# 				# "Propane": 0.1,
	# 			}

	# components = {
	# 	"Nitrogen": 0.0002,
	# 	"CO2": 0.0162,
	# 	"Methane": 0.2686,
	# 	"Ethane": 0.1445,
	# 	"Propane": 0.1240,
	# 	"i-Butane": 0.0262,
	# 	"n-Butane": 0.0490,
	# 	"i-Pentane": 0.0307,
	# 	"n-Pentane": 0.0294,
	# 	"n-Hexane": 0.0395,
	# 	"n-Heptane": 0.0614,
	# 	"n-Octane": 0.0598,
	# 	"n-Nonane": 0.0283,
	# 	"n-Decane": 0.0254,
	# 	"n-C11": 0.0154,
	# 	"n-C12": 0.0137,
	# 	"n-C13": 0.0121,
	# 	"n-C14": 0.0113,
	# 	"n-C15": 0.0089,
	# 	"n-C16": 0.0065,
	# 	"n-C17": 0.0057,
	# 	"n-C18": 0.0049,
	# 	"n-C19": 0.0040,
	# 	"n-C20": 0.0032,
	# 	"n-C21": 0.0024,
	# 	"n-C22": 0.0016,
	# 	"n-C23": 0.0016,
	# 	"n-C24": 0.0008,
	# 	"n-C25": 0.0008,
	# 	"n-C26": 0.0008,
	# 	"n-C27": 0.0008,
	# 	"n-C28": 0.0008,
	# 	"n-C29": 0.0008,
	# 	"n-C30": 0.0008,
	# }

	# C25+
	components = {
		"Nitrogen": 0.0026,
		"CO2": 0.0131,
		"Methane": 0.8466,
		"Ethane": 0.0632,
		"Propane": 0.0226,
		"i-Butane": 0.0037,
		"n-Butane": 0.0066,
		"i-Pentane": 0.0039,
		"n-Pentane": 0.0037,
		"n-Hexane": 0.0049,
		"n-Heptane": 0.0076,
		"n-Octane": 0.0074,
		"n-Nonane": 0.0035,
		"n-Decane": 0.0031,
		"n-C11": 0.0019,
		"n-C12": 0.0008,
		"n-C13": 0.0008,
		"n-C14": 0.0014,
		"n-C15": 0.0011,
		"n-C16": 0.00035,
		"n-C17": 0.00035,
		"n-C18": 0.000167,
		"n-C19": 0.000167,
		"n-C20": 0.000166,
		"n-C21": 0.00005,
		"n-C22": 0.00005,
		"n-C23": 0.00005,
		"n-C24": 0.00005,
		"n-C25": 0,
		"n-C26": 0,
		"n-C27": 0,
		"n-C28": 0,
		"n-C29": 0.0001,
		"n-C30": 0,
	}

	# C10+
	# components = {
	# 	"Nitrogen": 0.0026,
	# 	"CO2": 0.0131,
	# 	"Methane": 0.8466,
	# 	"Ethane": 0.0632,
	# 	"Propane": 0.0226,
	# 	"i-Butane": 0.0037,
	# 	"n-Butane": 0.0066,
	# 	"i-Pentane": 0.0039,
	# 	"n-Pentane": 0.0037,
	# 	"n-Hexane": 0.0049,
	# 	"n-Heptane": 0.0076,
	# 	"n-Octane": 0.0074,
	# 	"n-Nonane": 0.0035,
	# 	"n-Decane": 0.0106,
	# }

	di = DirectIntegration(15000, 75, 25000, 400, components, N = 10)

	# now = time.time()
	# print(di.calculate_correct_eta())
	# print(time.time() - now)
	# print(TOTAL_NUMBER)
	TOL = 1e-8
	print(di.calculate_outlet_temp(0.75))

	# di.N = 10
	# print(di.calculate_outlet_temp(0.75))

	# di.N = 1000
	# print(di.calculate_outlet_temp(0.75))