# -*- coding: utf-8 -*-

from __future__ import division

import win32com.client

HYSYS_EMPTY_CONST = -32767

EOS_LIST = [
				"pengrob",
				"BWRS",
		   ]

class HysysStream(object):
	"""A class to use the equation of state in Hysys through COM objects.
	"""

	def __init__(self, components, eos_type):
		"""Constructor of the HysysStream class.

		:param components: a dictionary of components and their mole fractions.
		:param eos_type: the EoS type to use. E.g.: "pengrob" or "BWRS"

		"""

		if eos_type not in EOS_LIST:
			raise ValueError("The EoS type: {0} is not a valid EoS type in Hysys.".format(eos_type))

		if abs(sum(components.values()) - 1) > 0.001:
			raise ValueError("The mole fractions (={0}) in the components dictionary must equal 1.".format(sum(components.values())))

		self.components = components
		self.eos_type = eos_type

		# initiate the pressure and temperature variables.
		self.pressure = HYSYS_EMPTY_CONST
		self.temperature = HYSYS_EMPTY_CONST
		self._enthalpy = HYSYS_EMPTY_CONST
		self.entropy = HYSYS_EMPTY_CONST

		self.initiate_hysys()

	def __del__(self):
		if hasattr(self, "sever"):
			self.server.Quit()

	def initiate_hysys(self):
		"""Create a new hysys case and add components from the components dict.
		and set the right EoS (fluid package).

		"""

		self.server = win32com . client . Dispatch ( "Hysys.Application" )
		self.case  = self.server.SimulationCases.Add()

		# start doing changes to the Basis Manager
		self.case.BasisManager.StartBasisChange()

		# add a Fluid Package
		self.case.BasisManager.FluidPackages.Add()

		# add the EoS Type
		self.case.BasisManager.FluidPackages(0).PropertyPackageName = self.eos_type

		# add all the components
		for component in self.components.keys():
			try:
				self.case.BasisManager.FluidPackages(0).ComponentList.Components.Add(component)
			except:
				raise ValueError("{0} is not a valid component in Hysys.".format(component))

		self.case.BasisManager.EndBasisChange()
		self.case.Flowsheet.Streams.Add()
		self.case.Flowsheet.Streams(0).ComponentMolarFractionValue = tuple(self.components.values())

	def set_pressure_value(self, pressure):
		"""
		Set a new pressure value to the stream. Will check if it's the stream is over specified
		and raise an exception if it is.

		:param pressure: the new pressure in kPa.
		:returns: True if successfully updated the pressure value.

		"""
		if self.pressure != pressure:
			try:
				self.case.Flowsheet.Streams(0).PressureValue = pressure
				self.pressure = pressure
			except:
				raise AttributeError("The stream is over-specified, therefore the pressure value can't be updated.")

		return True

	def set_temperature_value(self, temperature):
		"""
		Set a new temperature value to the stream. Will check if it's the stream is over specified
		and raise an exception if it is.

		:param temperature: the new temperature value in degrees Celsius.
		:returns: True if successfully updated the temperature value.

		"""
		if self.temperature != temperature:
			try:
				self.case.Flowsheet.Streams(0).TemperatureValue = temperature
				self.temperature = temperature
			except:
				raise AttributeError("The stream is over-specified, therefore the temperature value can't be updated.")

		return True

	def set_enthalpy_value(self, enthalpy):
		"""
		Set a new enthalpy value (molar enthalpy in kJ/kmole) to the the stream. Will check if it's the stream is over specified
		and raise an exception if it is.

		:params enthalpy: the new molar enthalpy value in kJ/kmole.
		:returns: True if successfully updated the enthalpy value.

		"""
		if self._enthalpy != enthalpy:
			try:
				self.case.Flowsheet.Streams(0).MolarEnthalpyValue = enthalpy
				self.enthalpy = enthalpy
			except:
				raise AttributeError("The stream is over-specified, therefore the enthalpy value can't be updated.")

	def update_pressure_and_temperature(self, pressure, temperature):
		"""Updates the pressure and temperature only if the values have changed.

		:param pressure: the new pressure in kPa.
		:param temperature: the new temperature in degrees C.

		"""
		if pressure != self.pressure or temperature != self.temperature:
			self.pressure = pressure
			self.temperature = temperature
			self.case.Flowsheet.Streams(0).PressureValue = self.pressure
			self.case.Flowsheet.Streams(0).TemperatureValue = self.temperature

	def enthalpy(self, pressure, temperature):
		"""Get enthalpy of the stream from a given pressure and temperature.

		:param pressure: the pressure in kPa
		:param temperature: the temperature in degrees Celsius.
		:returns: the enthalpy in kJ/kMole.

		"""
		# set pressure and temperature
		self.update_pressure_and_temperature(pressure, temperature)
		
		# get enthalpy from Hysys
		enthalpy = self.case.Flowsheet.Streams(0).MolarEnthalpyValue

		# return the extracted value
		return enthalpy

	def specific_volume(self, pressure, temperature):
		"""Get the specific volume for the given pressure and temperature.

		:param pressure: the pressure in kPa.
		:param temperature: the temperature in degrees Celsius.
		:returns: the specific volume in kMole/kg.

		"""
		# set pressure and temperature
		self.update_pressure_and_temperature(pressure, temperature)
		
		# get enthalpy from Hysys
		density = self.case.Flowsheet.Streams(0).MolarDensityValue

		# return the extracted value as specific volume
		return 1/density


if __name__ == '__main__':
	import time

	components = {
					"Propane": 0.1,
					"Methane": 0.2,
					"Ethane": 0.7,
				}

	eos = HysysStream(components, "pengrob")
	#print(eos.enthalpy(100, 25))
